/*
 * Copyright (C) 2024 sillysagiri
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
 */
 
#pragma once

#include "stb_image.h"
#include <cstdint>
#include <filesystem>
#include <string>
#include <vector>

#pragma pack(push, 1)  // Set the alignment to 1 byte (no padding)
struct Metadata {
	uint64_t header = 0x676D69796C6C6973;
	uint8_t version = 0;
	uint8_t format;
	uint8_t paletteId;
	uint16_t width;
	uint16_t height;
	uint8_t compression;
	uint32_t length;
};
#pragma pack(pop)  // Restore the previous alignment

enum Format
{
	Format_RGB_256 = 0,
	Format_RGB_16,
	Format_INDEXED_4,
	Format_INDEXED_16,
	Format_INDEXED_256,
	Format_INDEXED_32A3,
	Format_INDEXED_8A5,
	Format_PALETTE_16,
};

enum Compress
{
	Compress_NONE = 0,
	Compress_LZSS_WRAM,
	Compress_LZSS_VRAM,
	Compress_GZIP,
};

struct Image {
	int width, height, comp;
	stbi_uc *data;

	constexpr Image() : width(0), height(0), comp(0), data(nullptr) {}
	Image(const std::filesystem::path &path);
	~Image();

	void Load(const std::filesystem::path &path);
};

struct IndexedImage
{
	int width, height;
	uint8_t *data;

	constexpr IndexedImage() : width(0), height(0), data(nullptr) {}
	IndexedImage(int width, int height);
	~IndexedImage();

	void Load(int width, int height);
};

struct Color {
	uint8_t a;
	uint8_t r;
	uint8_t g;
	uint8_t b;
};

struct Params {
	std::vector<std::filesystem::path> input_list;
	std::filesystem::path output_dir;
	
	Format format;
	Compress compress;

	int paletteId;
	int limitColor = 0;
	bool verbose = false;
	bool paletteOnly = false;

	void Parse(int argc, char **argv);
	void Convert();
	void PrintHelp();
	void Print(const std::string &msg);

};

uint16_t RGB16(uint8_t a, uint8_t r, uint8_t g, uint8_t b);
void GetIndexed(const int numColor, std::vector<Image> &images, std::vector<Color> &palette, std::vector<IndexedImage> &indexes);
void BitPacking(const uint8_t bpp, IndexedImage &img, uint32_t &length);
void WritePalette(const std::string &path, const uint8_t paletteId, std::vector<Color> &palette);
void WriteImage(const std::string &path, const Format &format, const uint16_t width, const uint16_t height, const Compress &compress, const uint8_t paletteId, void *buffer, const uint32_t length);